'''
validates parsing
'''
import re
from datetime import date

from nepali.char import EnglishChar, nepali_to_english_text
from nepali.datetime import nepalidatetime


__nepali_time_re__CACHE = None

class NepaliTimeRE(dict):
    
    def __init__(self):
        """Create keys/values.
        Order of execution is important for dependency reasons.
        """
        base = super()
        base.__init__({
            # The " [1-9]" part of the regex is to make %c from ANSI C work
            'd': r"(?P<d>3[0-1]|[1-2]\d|0[1-9]|[1-9]| [1-9])",
            '-d': r"(?P<d>3[0-1]|[1-2]\d|0[1-9]|[1-9]| [1-9])", # same as "d"
            'f': r"(?P<f>[0-9]{1,6})",
            'H': r"(?P<H>2[0-3]|[0-1]\d|\d)",
            '-H': r"(?P<H>2[0-3]|[0-1]\d|\d)",
            'I': r"(?P<I>1[0-2]|0[1-9]|[1-9])",
            '-I': r"(?P<I>1[0-2]|0[1-9]|[1-9])",
            'G': r"(?P<G>\d\d\d\d)",
            'j': r"(?P<j>36[0-6]|3[0-5]\d|[1-2]\d\d|0[1-9]\d|00[1-9]|[1-9]\d|0[1-9]|[1-9])",
            'm': r"(?P<m>1[0-2]|0[1-9]|[1-9])",
            '-m': r"(?P<m>1[0-2]|0[1-9]|[1-9])", # same as "m"
            'M': r"(?P<M>[0-5]\d|\d)",
            '-M': r"(?P<M>[0-5]\d|\d)", # same as "M"
            'S': r"(?P<S>6[0-1]|[0-5]\d|\d)",
            '-S': r"(?P<S>6[0-1]|[0-5]\d|\d)", # same as "S"
            'w': r"(?P<w>[0-6])",

            'y': r"(?P<y>\d\d)",
            'Y': r"(?P<Y>\d\d\d\d)",
            'z': r"(?P<z>[+-]\d\d:?[0-5]\d(:?[0-5]\d(\.\d{1,6})?)?|(?-i:Z))",

            'A': self.__seqToRE(EnglishChar.days, 'A'),
            'a': self.__seqToRE(EnglishChar.days_half, 'a'),
            'B': self.__seqToRE(EnglishChar.months, 'B'),
            'b': self.__seqToRE(EnglishChar.months, 'b'),
            'p': self.__seqToRE(('AM', 'PM',), 'p'),

            '%': '%'
        })
    
    def __seqToRE(self, to_convert, directive):
        """Convert a list to a regex string for matching a directive.
        Want possible matching values to be from longest to shortest.  This
        prevents the possibility of a match occurring for a value that also
        a substring of a larger value that should have matched (e.g., 'abc'
        matching when 'abcdef' should have been the match).
        """
        to_convert = sorted(to_convert, key=len, reverse=True)
        for value in to_convert:
            if value != '':
                break
        else:
            return ''
        regex = '|'.join(re.escape(stuff) for stuff in to_convert)
        regex = '(?P<%s>%s' % (directive, regex)
        return '%s)' % regex

    def pattern(self, format):
        '''
        Handle conversion from format directives to regexes.
        '''
        processed_format = ''
        regex_chars = re.compile(r"([\\.^$*+?\(\){}\[\]|])")
        format = regex_chars.sub(r"\\\1", format)
        whitespace_replacement = re.compile(r'\s+')
        format = whitespace_replacement.sub(r'\\s+', format)
        while '%' in format:
            directive_index = format.index('%')+1
            index_increment = 1
            
            if format[directive_index] == '-':
                index_increment = 2
            
            if format[directive_index: directive_index+index_increment] not in self:
                return None

            processed_format = "%s%s%s" % (processed_format,
                                            format[:directive_index-1],
                                            self[format[directive_index: directive_index+index_increment]])
            format = format[directive_index+index_increment:]
        return "^%s%s$" % (processed_format, format)

    def compile(self, format):
        """Return a compiled re object for the format string."""
        return re.compile(self.pattern(format), re.IGNORECASE)

def get_nepali_time_re_object():
    global __nepali_time_re__CACHE
    if __nepali_time_re__CACHE == None:
        __nepali_time_re__CACHE = NepaliTimeRE()
    return __nepali_time_re__CACHE


def extract(datetime_str, format):
    '''
    Extracts year, month, day, hour, minute, etc from the given format.
    
    eg.
    USAGE: extract("2078-01-12", format="%Y-%m-%d")
    INPUT:
    datetime_str="2078-01-12" 
    format="%Y-%m-%d"

    OUTPUT:
    {
        "Y": 2078,
        "m": 1,
        "d": 12,
    }
    '''

    # converting datetime_str to english if any exists
    datetime_str = nepali_to_english_text(datetime_str)

    re_compiled_format = get_nepali_time_re_object().compile(format=format)
    match = re_compiled_format.match(datetime_str)
    if match == None:
        return {}
    return match.groupdict()

def transform(data: dict):
    '''
    transforms different format data to uniform data

    eg.
    INPUT:
    data = {
        "Y": 2078,
        "b": "Mangsir",
        "d": 12,
        ...
    }

    OUTPUT:
    {
        "year": 2078,
        "month": 8,
        "day": 12,
        ...
    }
    '''

    year = None
    month = day = 1
    hour = minute = second = fraction = 0
    weekday = None

    for date_key in data.keys():
        if date_key == 'y':
            year = int(data['y'])
            year += 2000
        elif date_key == 'Y':
            year = int(data['Y'])
        elif date_key == 'm':
            month = int(data['m'])
        elif date_key == 'B':
            # TODO: change indexing process
            month = EnglishChar.months.index(data['B'].lower().capitalize()) + 1
        elif date_key == 'b':
            # TODO: change indexing process
            month = EnglishChar.months.index(data['b'].lower().capitalize()) + 1
        elif date_key == 'd':
            day = int(data['d'])
        elif date_key == 'H':
            hour = int(data['H'])
        elif date_key == 'I':
            hour = int(data['I'])
            ampm = data.get('p', '').lower()
            # If there was no AM/PM indicator, we'll treat this like AM
            if ampm in ('', 'am'):
                # We're in AM so the hour is correct unless we're
                # looking at 12 midnight.
                # 12 midnight == 12 AM == hour 0
                if hour == 12:
                    hour = 0
            elif ampm == 'pm':
                # We're in PM so we need to add 12 to the hour unless
                # we're looking at 12 noon.
                # 12 noon == 12 PM == hour 12
                if hour != 12:
                    hour += 12
        elif date_key == 'M':
            minute = int(data['M'])
        elif date_key == 'S':
            second = int(data['S'])
        elif date_key == 'f':
            s = data['f']
            # Pad to always return microseconds.
            s += "0" * (6 - len(s))
            fraction = int(s)
        elif date_key == 'A':
            # TODO: change indexing process
            weekday = EnglishChar.days.index(data['A'].lower().capitalize()) + 1
        elif date_key == 'a':
            # TODO: change indexing process
            weekday = EnglishChar.days_half.index(data['a'].lower().capitalize()) + 1
        elif date_key == 'w':
            weekday = int(data['w'])
            if weekday == 0:
                weekday = 6
            else:
                weekday -= 1
    return {
        'year': year,
        'month': month,
        'day': day,
        'hour': hour,
        'minute': minute,
        'second': second,
        'microsecond': fraction,
        'weekday': weekday,
    }


def validate(datetime_str, format):
    '''
    validates datetime_str with the format
    Perform step by step test for fast performance. The steps are:
    -
    -
    returns False if validation failed
    returns nepalidatetime object if validation success.
    '''
    
    # 1. validate if format is correct.
    if get_nepali_time_re_object().pattern(format=format) == None:
        # regex pattern generation failed
        return None
    
    # 2. validate if parse result if not empty
    parsed_result = extract(datetime_str, format)
    if parsed_result.get('Y') == None and parsed_result.get('y') == None:
        # compilation failed or year included 
        return None

    # 3. validate if transformation
    transformed_data = transform(parsed_result)
    if transformed_data.get('year') == None:
        # could not transform data, not getting year
        return None

    # 4. create the datetime object
    return nepalidatetime(
        year=transformed_data['year'], 
        month=transformed_data['month'], 
        day=transformed_data['day'], 
        hour=transformed_data['hour'],
        minute=transformed_data['minute'],
        second=transformed_data['second'],
        microsecond=transformed_data['microsecond'],
    )


"""
Exceptions for nepali
"""

class InvalidDateFormatException(Exception):
    pass

class InvalidDateTimeFormatException(Exception):
    pass

class InvalidNepaliDateTimeObjectException(Exception):
    pass

class FormatNotMatchException(Exception):
    '''
    raised while parsing nepalidatetime format
    '''
    pass


from ._datetime import nepalidate, nepalitime, nepalidatetime, NepaliDate, NepaliDateTime, NepaliTime
from ._formarter import NepaliDateTimeFormater
from ._humanize import HumanizeDateTime, nepalihumanize

__all__ = [
	'nepalidate',
	'nepalitime',
	'nepalidatetime',

	'NepaliDate',
	'NepaliTime',
	'NepaliDateTime',

	'nepalihumanize',
	'NepaliDateTimeFormater',
	'HumanizeDateTime',
]


import datetime
import pytz
import warnings

from .timezone import NepaliTimeZone, get_timezone
from .exceptions import InvalidNepaliDateTimeObjectException

def to_utc_timezone(datetime_obj):
	if type(datetime_obj) != datetime.datetime:
		# None datetime object
		return datetime_obj

	if (not hasattr(datetime_obj, 'tzinfo')) or (not datetime_obj.tzinfo):
		datetime_obj = datetime_obj.replace(tzinfo=get_timezone())
	return datetime_obj.astimezone(pytz.timezone('UTC'))

def to_utc(datetime_obj):
	warnings.warn(
		message="to_utc is depreciated and no longer be available in version >= 1.0.0, use to_utc_timezone instead.", 
		category=DeprecationWarning
	)
	return to_utc_timezone(datetime_obj)

def to_nepali_timezone(datetime_obj):
	if type(datetime_obj) != datetime.datetime:
		return datetime_obj

	if (not hasattr(datetime_obj, 'tzinfo')) or (not datetime_obj.tzinfo):
		datetime_obj = datetime_obj.replace(tzinfo=get_timezone())
	return datetime_obj.astimezone(NepaliTimeZone())

def to_local(datetime_obj):
	warnings.warn(
		message="to_local is depreciated and no longer be available in version >= 1.0.0, use to_nepali_timezone instead.", 
		category=DeprecationWarning
	)
	return to_nepali_timezone(datetime_obj)

def to_nepalidatetime(datetime_object):
	"""
	Converts nepalidate, datetime.datetime, datetime.date to nepalidatetime. 
	"""
	from .datetime import nepalidate, nepalidatetime

	if type(datetime_object) == nepalidatetime:
		return datetime_object
	elif type(datetime_object) == nepalidate:
		return nepalidatetime.from_nepali_date(datetime_object)
	elif type(datetime_object) == datetime.datetime:
		return nepalidatetime.from_datetime(datetime_object)
	elif type(datetime_object) == datetime.date:
		return nepalidatetime.from_date(datetime_object)
	elif datetime_object == '' or datetime_object == None:
		return None
	raise InvalidNepaliDateTimeObjectException('Argument must be instance of nepalidate or nepalidatetime or datetime.datetime or datetime.date') 

def to_nepali_datetime(datetime_object):
	warnings.warn(
		message="to_nepali_datetime is depreciated and no longer be available in version >= 1.0.0, use to_nepalidatetime instead.", 
		category=DeprecationWarning
	)
	return to_nepalidatetime(datetime_object)

def to_nepalidate(datetime_object):
	nepalidatetime_obj = to_nepalidatetime(datetime_object)
	if nepalidatetime_obj != None:
		return nepalidatetime_obj.date()
	return 


import datetime as pythonDateTime
from nepali.char import EnglishChar, NepaliChar

from nepali.exceptions import InvalidDateFormatException, InvalidNepaliDateTimeObjectException

from ._datetime import nepalidate, nepalidatetime

class NepaliDateTimeFormater:
	"""
	NepaliDateTimeFormater: formats nepali datetime to string ( using strftime )
	"""

	# format according to python's datetime with class method
	format_map = {
		'a': 'weekdayHalf',
		'A': 'weekdayFull',
		'w': 'weekdayNumber',
		'd': 'day',
		'-d': 'day_nonzero',
		'b': 'monthFull',
		'B': 'monthFull',
		'm': 'monthNumber',
		'-m': 'monthNumber_nonzero',
		'y': 'yearHalf',
		'Y': 'yearFull',
		'H': 'hour24',
		'-H': 'hour24_nonzero',
		'I': 'hour12',
		'-I': 'hour12_nonzero',
		'p': 'ampm',
		'M': 'minute',
		'-M': 'minute_nonzero',
		'S': 'second',
		'-S': 'second_nonzero',
	}

	def __init__(self, datetime_object, english=False):
		if type(datetime_object) == nepalidatetime:
			self.npDateTime = datetime_object
		elif type(datetime_object) == nepalidate:
			self.npDateTime = datetime_object.to_nepalidatetime()
		elif type(datetime_object) == pythonDateTime.date:
			self.npDateTime = nepalidatetime.from_date(datetime_object)
		elif type(datetime_object) == pythonDateTime.datetime:
			self.npDateTime = nepalidatetime.from_datetime(datetime_object)
		else:
			raise InvalidNepaliDateTimeObjectException('Argument must be instance of nepalidate or nepalidatetime or datetime.datetime or datetime.date') 


		self.english = english

	def __str__(self):
		return str(self.npDateTime)

	def get_str(self, format):
		""" generates formated string """
		i, n = 0, len(format)
		time_str = []
		try:
			while i < n:
				ch = format[i]
				i += 1
				if ch == '%':
					if i < n:
						ch = format[i]
						
						if ch == '%':
							# for % character
							time_str.append('%')

						elif ch == '-':
							# special mid characters eg. "-" for non zero padded values
							special_ch = ch
							if i+1 < n:
								i += 1
								ch = format[i]
								time_str.append(getattr(self, self.get_format_map(special_ch+ch)))
						else:
							# mapping % forwarded character
							time_str.append(getattr(self, self.get_format_map(ch)))
						i += 1
				else:
					time_str.append(ch)
		except InvalidDateFormatException as e:
			raise e
		except Exception:
			raise Exception('Unable to convert NepaliDateTime to str')
		time_str = ''.join(time_str)

		return time_str

	def get_format_map(self, ch):
		if ch not in self.format_map:
			raise InvalidDateFormatException('Invalid Date format %{}'.format(ch))
		return self.format_map.get(ch)

	@property
	def weekdayHalf(self):
		"""
		%a
		"""
		if self.english:
			return EnglishChar.half_day(self.npDateTime.week_day)
		return NepaliChar.half_day(self.npDateTime.week_day)

	@property
	def weekdayFull(self):
		"""
		%A
		"""
		if self.english:
			return EnglishChar.day(self.npDateTime.week_day)
		return NepaliChar.day(self.npDateTime.week_day)

	@property
	def weekdayNumber(self):
		"""
		%w
		"""
		if self.english:
			return str(self.npDateTime.week_day-1)
		return NepaliChar.number(self.npDateTime.week_day-1)

	@property
	def day(self):
		"""
		%d
		"""
		day = str(self.npDateTime.day)
		if len(day) < 2:
			day = '0'+day
		if self.english:
			return str(day)
		return NepaliChar.number(day)

	@property
	def day_nonzero(self):
		"""
		%D
		"""
		day = str(self.npDateTime.day)
		if self.english:
			return str(day)
		return NepaliChar.number(day)

	@property
	def monthFull(self):
		"""
		%B or %b
		"""
		if self.english:
			return EnglishChar.month(self.npDateTime.month)
		return NepaliChar.month(self.npDateTime.month)

	@property
	def monthNumber(self):
		"""
		%m
		"""
		month = str(self.npDateTime.month)
		if len(month) < 2:
			month = '0'+month
		if self.english:
			return str(month)
		return NepaliChar.number(month)

	@property
	def monthNumber_nonzero(self):
		"""
		%m
		"""
		month = str(self.npDateTime.month)
		if self.english:
			return str(month)
		return NepaliChar.number(month)

	@property
	def yearHalf(self):
		"""
		%y
		"""
		if self.english:
			return str(self.npDateTime.year)[2:]
		return NepaliChar.number(str(self.npDateTime.year)[2:])

	@property
	def yearFull(self):
		"""
		%Y
		"""
		if self.english:
			return str(self.npDateTime.year)
		return NepaliChar.number(self.npDateTime.year)
	
	@property
	def hour24(self):
		"""
		%H
		"""
		hour = str(self.npDateTime.hour)
		if len(hour) < 2:
			hour = '0'+hour
		if self.english:
			return str(hour)
		return NepaliChar.number(hour)

	@property
	def hour24_nonzero(self):
		"""
		%H
		"""
		hour = self.npDateTime.hour
		if self.english:
			return str(hour)
		return NepaliChar.number(hour)
	
	@property
	def hour12(self):
		"""
		%I
		"""
		hour = self.npDateTime.hour
		if hour > 12:
			hour = hour - 12
		if hour == 0:
			hour = 12
		hour = str(hour)
		if len(hour) < 2:
			hour = '0'+hour

		if self.english:
			return str(hour)
		return NepaliChar.number(hour)

	@property
	def hour12_nonzero(self):
		"""
		%I
		"""
		hour = self.npDateTime.hour
		if hour > 12:
			hour = hour - 12
		if hour == 0:
			hour = 12
		hour = str(hour)
		if self.english:
			return str(hour)
		return NepaliChar.number(hour)
	
	@property
	def ampm(self):
		"""
		%p
		"""
		if self.english:
			ampm = 'AM'
			if self.npDateTime.hour > 12:
				ampm = 'PM'
			return str(ampm)

		ampm = ''
		if self.npDateTime.hour < 12:
			ampm = 'शुभप्रभात'
		elif self.npDateTime.hour >= 12 and self.npDateTime.hour < 18:
			ampm = 'मध्यान्ह'
		else:
			ampm = 'अपरान्ह'
		return str(ampm)
	
	@property
	def minute(self):
		"""
		%M
		"""
		minute = str(self.npDateTime.minute)
		if len(minute) < 2:
			minute = '0'+minute
		if self.english:
			return str(minute)
		return NepaliChar.number(minute)

	@property
	def minute_nonzero(self):
		"""
		%M
		"""
		minute = str(self.npDateTime.minute)
		if self.english:
			return str(minute)
		return NepaliChar.number(minute)
	
	@property
	def second(self):
		"""
		%S
		"""
		second = str(self.npDateTime.second)
		if len(second) < 2:
			second = '0'+second
		if self.english:
			return str(second)
		return NepaliChar.number(second)

	@property
	def second_nonzero(self):
		"""
		%S
		"""
		second = str(self.npDateTime.second)
		if self.english:
			return str(second)
		return NepaliChar.number(second)





from django import template
from django.utils import timezone

from nepali.datetime import nepalidatetime, nepalihumanize as humanize
from nepali.utils import to_local, to_utc, to_nepali_datetime

register = template.Library()

@register.filter(name='nepalidate')
def nepalidate(datetime_obj, format="%B %d, %Y, %A"):
	nepali_datetime_obj = to_nepali_datetime(datetime_obj)
	if nepali_datetime_obj == None: return None
	return nepali_datetime_obj.strftime(format)

@register.filter(name='nepalidate_en')
def nepalidate_en(datetime_obj, format="%B %d, %Y, %A"):
	nepali_datetime_obj = to_nepali_datetime(datetime_obj)
	if nepali_datetime_obj == None: return None
	return nepali_datetime_obj.strftime_en(format)

@register.filter(name='nepalihumanize')
def nepalihumanize(datetime_obj, threshold=None, format=None):
	""" templatetag to humanize nepalidatetime """
	nepali_datetime_obj = to_nepali_datetime(datetime_obj)
	if nepali_datetime_obj == None: return None
	return humanize(nepali_datetime_obj, threshold=threshold, format=format)
	
@register.simple_tag
def nepalinow(format="%B %d, %Y, %A"):
	""" templatetag to display current datetime in nepali format """
	return to_nepali_datetime(timezone.now()).strftime(format)



import datetime
import unittest

from nepali.datetime import nepalidate, nepalitime, nepalidatetime
from nepali.utils import to_nepali_timezone

class TestNepaliDateTime(unittest.TestCase):

    #
    # nepalidate

    def test_nepalidate_now_and_today(self):
        today = nepalidate.today()
        now = nepalidate.now()
        self.assertEqual(today, now, msg='nepalidate today is not equals to now.')

    def test_nepalidate_comparision(self):
        nepalidate_obj1 = nepalidate(year=2051, month=4, day=28)

        # test equal date
        nepalidate_obj2 = nepalidate(year=2051, month=4, day=28)
        self.assertEqual(nepalidate_obj1, nepalidate_obj2, 'Same nepali date are not checked equals.')

        # test different date
        nepalidate_obj3 = nepalidate(year=2051, month=4, day=29)
        self.assertNotEqual(nepalidate_obj1, nepalidate_obj3, 'Different nepali date are checked as equal.')

        # TODO: check other comparisions (later)

    def test_nepalidate_year_month_day(self):
        nepalidate_obj = nepalidate(year=2051, month=4, day=28)
        self.assertEqual(nepalidate_obj.year, 2051)
        self.assertEqual(nepalidate_obj.month, 4)
        self.assertEqual(nepalidate_obj.day, 28)
        self.assertEqual(nepalidate_obj.weekday(), 4)

    def test_nepalidate_from_date(self):
        python_date_obj = datetime.date(year=1994, month=8, day=12)
        nepalidate_obj = nepalidate.from_date(python_date_obj)
        self.assertEqual(python_date_obj, nepalidate_obj.to_date(), 'nepalidate from_date and to_date are not equal.')

    def test_nepalidate_from_datetime(self):
        python_datetime_obj = datetime.datetime(year=1994, month=8, day=12, hour=4, minute=8, second=12)
        nepalidate_obj = nepalidate.from_datetime(python_datetime_obj)
        self.assertEqual(python_datetime_obj.date(), nepalidate_obj.to_date(), 'nepalidate from_datetime and to_datetime are not equal.')

    def test_nepalidate_strftime(self):
        nepalidate_obj = nepalidate(year=2051, month=4, day=28)
        self.assertEqual(nepalidate_obj.strftime('%Y-%m-%d'), '२०५१-०४-२८')
        self.assertEqual(nepalidate_obj.strftime('%a %A %w %d %b %B %m %y %Y'), 'बुध बुधबार ३ २८ साउन साउन ०४ ५१ २०५१')
        self.assertEqual(nepalidate_obj.strftime_en('%Y-%m-%d'), '2051-04-28')
        self.assertEqual(nepalidate_obj.strftime_en('%a %A %w %d %b %B %m %y %Y'), 'Wed Wednesday 3 28 Sharwan Sharwan 04 51 2051')

    def test_nepalidate_strptime(self):
        nepalidatetime_obj = nepalidate.strptime('2078-01-12', format='%Y-%m-%d')
        self.assertEqual(nepalidatetime_obj.year, 2078)
        self.assertEqual(nepalidatetime_obj.month, 1)
        self.assertEqual(nepalidatetime_obj.day, 12)

    #
    # nepalitime
    
    def test_nepalitime_now(self):
        nepalitime_obj, dt_now = nepalitime.now(), datetime.datetime.now()
        self.assertEqual(nepalitime_obj.hour, dt_now.hour)
        self.assertEqual(nepalitime_obj.minute, dt_now.minute)
        self.assertEqual(nepalitime_obj.second, dt_now.second)

    #
    # nepalidatetime

    def test_nepalidatetime_from_date(self):
        python_datetime_obj = datetime.datetime(year=1994, month=8, day=12, hour=5, minute=28, second=23)
        nepalidatetime_obj = nepalidatetime.from_date(python_datetime_obj)
        self.assertEqual(python_datetime_obj.date(), nepalidatetime_obj.to_date())

    def test_nepalidatetime_from_datetime(self):
        python_datetime_obj = datetime.datetime(year=1994, month=8, day=12, hour=5, minute=28, second=23)
        nepalidatetime_obj = nepalidatetime.from_datetime(python_datetime_obj)
        self.assertEqual(to_nepali_timezone(python_datetime_obj), nepalidatetime_obj.to_datetime())

    def test_nepalidatetime_strftime(self):
        nepalidatetime_obj = nepalidatetime(year=2051, month=4, day=28, hour=5, minute=28, second=23)
        self.assertEqual(nepalidatetime_obj.strftime('%Y-%m-%d %H:%M:%S'), '२०५१-०४-२८ ०५:२८:२३')
        self.assertEqual(nepalidatetime_obj.strftime('%a %A %w %d %b %B %m %y %Y %H %I %p %M %S'), 'बुध बुधबार ३ २८ साउन साउन ०४ ५१ २०५१ ०५ ०५ शुभप्रभात २८ २३')
        self.assertEqual(nepalidatetime_obj.strftime_en('%Y-%m-%d %H:%M:%S'), '2051-04-28 05:28:23')
        self.assertEqual(nepalidatetime_obj.strftime_en('%a %A %w %d %b %B %m %y %Y %H %I %p %M %S'), 'Wed Wednesday 3 28 Sharwan Sharwan 04 51 2051 05 05 AM 28 23')

    def test_nepalidatetime_strptime(self):
        nepalidatetime_obj = nepalidatetime.strptime('2078-01-12 13:12', format='%Y-%m-%d %H:%M')
        self.assertEqual(nepalidatetime_obj.year, 2078)
        self.assertEqual(nepalidatetime_obj.month, 1)
        self.assertEqual(nepalidatetime_obj.day, 12)
        self.assertEqual(nepalidatetime_obj.hour, 13)
        self.assertEqual(nepalidatetime_obj.minute, 12)

    def test_nepalidatetime_timedelta(self):
        nepalidatetime_obj1 = nepalidatetime(year=2051, month=4, day=28, hour=5, minute=28, second=23)
        nepalidatetime_obj2 = nepalidatetime(year=2051, month=4, day=29, hour=5, minute=28, second=23)
        nepalidatetime_obj1 = nepalidatetime_obj1 + datetime.timedelta(days=1)
        self.assertEqual(nepalidatetime_obj1, nepalidatetime_obj2)



import unittest

from nepali.char import nepali_to_english_text
from nepali.datetime import nepalidatetime
from nepali.datetime import parser as nepalidatetime_parser
from nepali.datetime.parser.validators import NepaliTimeRE, extract, transform, validate
from nepali.exceptions import FormatNotMatchException, InvalidDateTimeFormatException


class TestNepaliDateTimeParserValidators(unittest.TestCase):
    '''
    Tests nepali date
    '''

    def setUp(self) -> None:
        self.nepali_time_re = NepaliTimeRE()
        return super().setUp()
    
    # nepali time re tests

    def test_format_to_regex(self):
        format_to_regex = self.nepali_time_re.pattern('%Y $ \d %-d *')
        self.assertEqual(format_to_regex, r'^(?P<Y>\d\d\d\d)\s+\$\s+\\d\s+(?P<d>3[0-1]|[1-2]\d|0[1-9]|[1-9]| [1-9])\s+\*$')

    def test_all_format_to_regex(self):
        format_to_regex = self.nepali_time_re.pattern('%a %A %w %d %b %B %m %y %Y %H %I %p %M %S')
        expected_output = r'^(?P<a>Sun|Mon|Tue|Wed|Thu|Fri|Sat)\s+(?P<A>Wednesday|Thursday|Saturday|Tuesday|Sunday|Monday|Friday)\s+(?P<w>[0-6])\s+(?P<d>3[0-1]|[1-2]\d|0[1-9]|[1-9]| [1-9])\s+(?P<b>Baishakh|Sharwan|Mangsir|Chaitra|Jestha|Bhadra|Ashwin|Kartik|Falgun|Ashad|Poush|Magh)\s+(?P<B>Baishakh|Sharwan|Mangsir|Chaitra|Jestha|Bhadra|Ashwin|Kartik|Falgun|Ashad|Poush|Magh)\s+(?P<m>1[0-2]|0[1-9]|[1-9])\s+(?P<y>\d\d)\s+(?P<Y>\d\d\d\d)\s+(?P<H>2[0-3]|[0-1]\d|\d)\s+(?P<I>1[0-2]|0[1-9]|[1-9])\s+(?P<p>AM|PM)\s+(?P<M>[0-5]\d|\d)\s+(?P<S>6[0-1]|[0-5]\d|\d)$'
        self.assertEqual(format_to_regex, expected_output)

    # test extract

    def test_simple_extract(self):
        extracted_data = extract("2078-01-12", format="%Y-%m-%d")
        self.assertEqual(extracted_data.get('Y'), '2078')
        self.assertEqual(extracted_data.get('m'), '01')
        self.assertEqual(extracted_data.get('d'), '12')

    def test_complex_extract(self):
        extracted_data = extract("Wed Wednesday 3 28 Sharwan Sharwan 04 51 2051 05 05 AM 28 23", format="%a %A %w %d %b %B %m %y %Y %H %I %p %M %S")
        self.assertEqual(extracted_data.get('a'), 'Wed')
        self.assertEqual(extracted_data.get('A'), 'Wednesday')
        self.assertEqual(extracted_data.get('w'), '3')
        self.assertEqual(extracted_data.get('d'), '28')
        self.assertEqual(extracted_data.get('b'), 'Sharwan')
        self.assertEqual(extracted_data.get('B'), 'Sharwan')
        self.assertEqual(extracted_data.get('m'), '04')
        self.assertEqual(extracted_data.get('y'), '51')
        self.assertEqual(extracted_data.get('Y'), '2051')
        self.assertEqual(extracted_data.get('H'), '05')
        self.assertEqual(extracted_data.get('I'), '05')
        self.assertEqual(extracted_data.get('p'), 'AM')
        self.assertEqual(extracted_data.get('M'), '28')
        self.assertEqual(extracted_data.get('S'), '23')

    def test_nepali_to_english_text_conversion(self):
        self.assertEqual(
            nepali_to_english_text('बुध बुधबार ३ २८ साउन साउन ०४ ५१ २०५१ ०५ ०५ शुभप्रभात २८ २३'),
            'Wed Wednesday 3 28 Sharwan Sharwan 04 51 2051 05 05 AM 28 23'
        )

    def test_complex_extract_in_nepali(self):
        extracted_data = extract("बुध बुधबार ३ २८ साउन साउन ०४ ५१ २०५१ ०५ ०५ शुभप्रभात २८ २३", format="%a %A %w %d %b %B %m %y %Y %H %I %p %M %S")
        self.assertEqual(extracted_data.get('a'), 'Wed')
        self.assertEqual(extracted_data.get('A'), 'Wednesday')
        self.assertEqual(extracted_data.get('w'), '3')
        self.assertEqual(extracted_data.get('d'), '28')
        self.assertEqual(extracted_data.get('b'), 'Sharwan')
        self.assertEqual(extracted_data.get('B'), 'Sharwan')
        self.assertEqual(extracted_data.get('m'), '04')
        self.assertEqual(extracted_data.get('y'), '51')
        self.assertEqual(extracted_data.get('Y'), '2051')
        self.assertEqual(extracted_data.get('H'), '05')
        self.assertEqual(extracted_data.get('I'), '05')
        self.assertEqual(extracted_data.get('p'), 'AM')
        self.assertEqual(extracted_data.get('M'), '28')
        self.assertEqual(extracted_data.get('S'), '23')


    # test transform
    def test_transform(self):
        data = transform({
            'Y': '2078',
            'b': 'MAngsir',
            'd': '12',
            'p': 'pm',
            'I': '05',
            'M': '25',
            'S': '31',
            'f': '455',
            'a': 'Wed',
        })
        self.assertEqual(data.get('year'), 2078)
        self.assertEqual(data.get('month'), 8)
        self.assertEqual(data.get('day'), 12)
        self.assertEqual(data.get('hour'), 17)
        self.assertEqual(data.get('minute'), 25)
        self.assertEqual(data.get('second'), 31)
        self.assertEqual(data.get('microsecond'), 455000)
        self.assertEqual(data.get('weekday'), 4)

    #
    # test validate

    def test_validate(self):
        nepalidatetime_obj = validate('2078-01-12', format='%Y-%m-%d')
        self.assertEqual(nepalidatetime_obj.year, 2078)
        self.assertEqual(nepalidatetime_obj.month, 1)
        self.assertEqual(nepalidatetime_obj.day, 12)

    def test_test_validate(self):
        nepalidatetime_obj = validate('29 Jestha, 2078, 1:30 PM', format='%d %B, %Y, %I:%M %p')
        self.assertEqual(nepalidatetime_obj.year, 2078)
        self.assertEqual(nepalidatetime_obj.month, 2)
        self.assertEqual(nepalidatetime_obj.day, 29)
        self.assertEqual(nepalidatetime_obj.hour, 13)
        self.assertEqual(nepalidatetime_obj.minute, 30)
        self.assertEqual(nepalidatetime_obj.second, 0)

    def test_validate_with_invalid_data(self):
        self.assertEqual(validate('2078-01-12', format='%k-%m-%d'), None)
        self.assertEqual(validate('2078-01-12', format='%m-%M-%Y'), None)


class TestNepaliDateTimeParser(unittest.TestCase):
    '''
    Tests nepali datetime parser.
    '''

    # test strptime
    def test_strptime(self):
        nepalidatetime_obj = nepalidatetime_parser.strptime('2078-01-12', format='%Y-%m-%d')
        self.assertEqual(nepalidatetime_obj.year, 2078)
        self.assertEqual(nepalidatetime_obj.month, 1)
        self.assertEqual(nepalidatetime_obj.day, 12)

    def test_strptime_fail(self):
        with self.assertRaises(FormatNotMatchException):
            nepalidatetime_parser.strptime('2078-01-12', format='%Y %d')

    # test parser

    def test_normal_string_parse(self):
        parsed_datetime = nepalidatetime_parser.parse('2071-01-24')
        test_datetime = nepalidatetime(2071, 1, 24)
        self.assertEqual(parsed_datetime, test_datetime)

    def test_complex_string_parse(self):
        nepalidatetime_obj = nepalidatetime_parser.parse('29 Jestha, 2078, 1:30 PM')
        self.assertEqual(nepalidatetime_obj.year, 2078)
        self.assertEqual(nepalidatetime_obj.month, 2)
        self.assertEqual(nepalidatetime_obj.day, 29)
        self.assertEqual(nepalidatetime_obj.hour, 13)
        self.assertEqual(nepalidatetime_obj.minute, 30)

    def test_parse_failed(self):
        with self.assertRaises(InvalidDateTimeFormatException):
            nepalidatetime_parser.parse('')



import datetime as pythonDateTime
from nepali.char import NepaliChar

from nepali.exceptions import InvalidNepaliDateTimeObjectException
from nepali.utils import to_nepali_timezone

from ._datetime import nepalidate, nepalidatetime

class HumanizeDateTime:
	"""
	HumanizeDate converts NepaliDateTime to nepali human readable form
	"""

	__past_text = "अघि"
	__future_text = "पछि"
	__now_text = "भर्खरै"
	__year_text = "वर्ष"
	__month_text = "महिना"
	__day_text = "दिन"
	__hour_text = "घण्टा"
	__minute_text = "मिनेट"
	__second_text = "सेकेन्ड"

	def	__init__(self, datetime_obj, *args, **kwargs):
		""" 
		initializes humanize class
		datetime_obj: python datetime object to be humanized
		threshold (kwargs): threshold to be humanize
		format (kwargs): format to display behind threshold
		"""
		if type(datetime_obj) == nepalidatetime:
			self.datetime_obj = datetime_obj.to_datetime()
		elif type(datetime_obj) == nepalidate:
			self.datetime_obj = nepalidatetime.from_nepali_date(datetime_obj).to_datetime()
		elif type(datetime_obj) == pythonDateTime.date:
			self.datetime_obj = nepalidatetime.from_date(datetime_obj).to_datetime()
		elif type(datetime_obj) == pythonDateTime.datetime:
			self.datetime_obj = to_nepali_timezone(datetime_obj)
		else:
			raise InvalidNepaliDateTimeObjectException('Argument must be instance of NepaliDate or NepaliDateTime or datetime.datetime or datetime.date') 

		self.threshold = kwargs.get('threshold')
		self.format = kwargs.get('format')

		# seconds after from now to datetime_obj
		self.seconds = None


	def __calc_seconds(self):
		""" calculates total seconds from now """

		current_date_time = now()
		date = self.datetime_obj
		self.seconds = int((current_date_time-date).total_seconds())
		self.interval_tense = self.__past_text
		if(self.seconds < 0):
			self.interval_tense = self.__future_text

	def to_str(self):
		""" returns humanize string """

		self.__calc_seconds()	# refreshing seconds
		seconds = self.seconds
		if( seconds < 0):
			seconds = 0 - seconds

		if not self.threshold == None:
			if( seconds >= self.threshold):
				return self.get_datetime().strip()
		
		return self.get_humanize().strip()


	def get_humanize(self):
		"""
		returns humanize datetime
		"""
		self.__calc_seconds()	# refreshing seconds

		interval_value = 0
		interval_text = ""
		if( self.seconds == 0 ):
			# now
			return self.__now_text

		elif( self.seconds < 60):
			# seconds
			interval_value = self.seconds
			interval_text = self.__second_text
			
		elif( self.seconds < 3600):
			# minute
			interval_value = self.seconds//60
			interval_text = self.__minute_text

		elif( self.seconds < 86400):
			# hour
			interval_value = self.seconds//3600
			interval_text = self.__hour_text

		elif( self.seconds < 2592000):
			# day
			interval_value = self.seconds//86400
			interval_text = self.__day_text

		elif( self.seconds < 946080000):
			# month
			interval_value = self.seconds//2592000
			interval_text = self.__month_text

		else:
			# year
			interval_value = self.seconds//946080000
			interval_text = self.__year_text

		interval_value = NepaliChar.number(interval_value)
		return str(interval_value)+' '+str(interval_text)+' '+self.interval_tense


	def get_datetime(self):
		"""
		returns date in nepali characters
		"""
		if not self.format:
			self.format = '%B %d, %Y'
		ndt = nepalidatetime.from_datetime(self.datetime_obj)
		return ndt.strftime(self.format)

	def __str__(self):
		return self.to_str()

	def __repr__(self):
		return str(self)


def nepalihumanize(datetime_obj, threshold=None, format=None):
	""" returns to humanize nepalidatetime """
	return HumanizeDateTime(datetime_obj, threshold=threshold, format=format).to_str()



import datetime

class NepaliTimeZone(datetime.tzinfo):
    def utcoffset(self, dt):
        return self.dst(dt) + datetime.timedelta(hours=5, minutes=45)

    def dst(self, dt):
        return datetime.timedelta(0)
    
    def tzname(self,dt):
        return "Asia/Kathmandu"

    def __str__(self):
        return "Asia/Kathmandu"

def get_timezone():
    return datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo

def now():
    return datetime.datetime.now(get_timezone())

def utc_now():
    return datetime.datetime.now(datetime.timezone.utc)




from nepali.exceptions import InvalidDateTimeFormatException, FormatNotMatchException

from.validators import validate

__all__ = ["strptime", "parse",]

_standard_datetime_format_CACHE = None

def strptime(datetime_str, format):
	'''
	Parses nepalidatetime str with the corresponding format.
	returns nepalidatetime object.
	'''
	nepalidatetime_object = validate(datetime_str, format)
	if nepalidatetime_object == None:
		raise FormatNotMatchException('Datetime string did not match with the given format.') 
	return nepalidatetime_object

def _get_standard_formats():
	global _standard_datetime_format_CACHE
	if _standard_datetime_format_CACHE != None:
		return _standard_datetime_format_CACHE

	STANDARD_DATE_FORMAT = [
		'%Y-%m-%d',
		'%d-%m-%Y',
		'%Y %m %d',
		'%Y %B %d',
		'%d %B %Y',
		'%d %B, %Y',
		'%B %d, %Y',
		'%Y/%m/%d',
		'%d/%m/%Y',
	]

	STANDARD_TIME_FORMAT = [
		'%H:%M',
		'%I:%M %p',
		'%I %p',
		'%I%p',
		'%H:%M:%S',
		'%H:%M:%S:%f',
	]

	_standard_datetime_format_CACHE = STANDARD_DATE_FORMAT.copy()
	for time_format in STANDARD_TIME_FORMAT:
		for date_format in STANDARD_DATE_FORMAT:
			_standard_datetime_format_CACHE += ['{} {}'.format(date_format, time_format)]
			_standard_datetime_format_CACHE += ['{}, {}'.format(date_format, time_format)]
	return _standard_datetime_format_CACHE


def parse(datetime_str):
	"""
	parses nepali datetime
	eg. parse('2078-10-12') => <NepaliDateTime: 2078-10-12>
	"""
	standard_formats = _get_standard_formats()
	nepalidatetime_object = None
	for format in standard_formats:
		nepalidatetime_object = validate(datetime_str, format=format)
		if nepalidatetime_object != None:
			return nepalidatetime_object

	raise InvalidDateTimeFormatException('Invalid format to parse nepali datetime.') 


class NepaliNumber:
	
	@classmethod
	def convert_and_add_comma(cls, number):
		return cls.add_comma(cls.convert(number))

	@staticmethod
	def convert(num):
		nepaliNumbers = ['०','१','२','३','४','५','६','७','८','९']
		np_num = ''
		en_num = str(num)
		for e in en_num:
			if e in ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']:
				np_num = np_num+str(nepaliNumbers[int(e)])
			else:
				np_num = np_num+str(e)
		return np_num

	@staticmethod
	def revert(num):
		nepaliNumbers = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']
		np_num = ''
		en_num = str(num)
		for e in en_num:
			if e in ['०','१','२','३','४','५','६','७','८','९']:
				np_num = np_num+str(nepaliNumbers[int(e)])
			else:
				np_num = np_num+str(e)
		return np_num

	@staticmethod
	def add_comma(number):
		number_with_comma = ""
		counter = 0
		for nepali_number_char in list(str(number))[::-1]:
			if counter == 3:
				number_with_comma = ',' + number_with_comma

			elif counter != 1 and counter != 3 and (counter-1) % 2 == 0:
				number_with_comma = ',' + number_with_comma
			
			number_with_comma = nepali_number_char + number_with_comma

			# increasing counter	
			counter += 1
		
		return number_with_comma

	@staticmethod
	def add_comma_english(number):
		return '{:,}'.format(number)



"""
setup file for nepali package

- Publish version
$ python setup.py publish

"""


import os
import sys
import setuptools

with open("README.md", "r") as fh:
	long_description = fh.read()


if sys.argv[-1] == 'publish':
	if os.system("pip3 freeze | grep twine"):
		print("twine not installed.\nUse `pip install twine`.\nExiting.")
		sys.exit()
	os.system("rm -rf dist")
	os.system("python3 setup.py sdist bdist_wheel")
	os.system("twine upload dist/*")
	sys.exit()


setuptools.setup(
	name="nepali",
	version="0.5.1",
	author="Ajesh Sen Thapa",
	author_email="aj3sshh@gmail.com",
	description="Python package for nepali processes",
	long_description=long_description,
	long_description_content_type="text/markdown",
	url="https://github.com/aj3sh/nepali",
	packages=setuptools.find_packages(),
	test_suite='nepali.tests',
	install_requires=[
		'pytz'
	],
	classifiers=[

		'Programming Language :: Python :: 3',
		'License :: OSI Approved :: MIT License',
		'Operating System :: OS Independent',
	
	],
)



from django import template

from nepali.number import NepaliNumber

register = template.Library()

@register.filter(name='nepalinumber')
def nepalinumber(value):
	return NepaliNumber.convert(value)

@register.filter(name='nepalinumber_with_comma')
def nepalinumber_with_comma(value):
	return NepaliNumber.convert_and_add_comma(value)


from .number import NepaliNumber


MONTHS_MAP = {
	'बैशाख': 'Baishakh',
	'जेठ': 'Jestha',
	'असार': 'Ashad',
	'साउन': 'Sharwan',
	'भदौ': 'Bhadra',
	'असोज': 'Ashwin',
	'कात्तिक': 'Kartik',
	'मंसिर': 'Mangsir',
	'पुस': 'Poush',
	'माघ': 'Magh',
	'फागुन': 'Falgun',
	'चैत': 'Chaitra',
}

DAYS_MAP = {
	'आइतबार': 'Sunday',
	'सोमबार': 'Monday',
	'मंगलबार': 'Tuesday',
	'बुधबार': 'Wednesday',
	'बिहीबार': 'Thursday',
	'शुक्रबार': 'Friday',
	'शनिबार': 'Saturday',
}

HALF_DAYS_MAP = {
	'आइत': 'Sun',
	'सोम': 'Mon',
	'मंगल': 'Tue',
	'बुध': 'Wed',
	'बिही': 'Thu',
	'शुक्र': 'Fri',
	'शनि': 'Sat',
}

AM_PM_MAP = {
	'शुभप्रभात' : 'AM',
	'मध्यान्ह': 'PM',
	'अपरान्ह': 'PM',
}



class NepaliChar:

	@staticmethod
	def number(num):
		return NepaliNumber.convert(num)

	@staticmethod
	def day(day):
		days = ['आइतबार','सोमबार','मंगलबार','बुधबार','बिहीबार','शुक्रबार','शनिबार']
		return days[day-1]

	@staticmethod
	def half_day(day):
		days = ['आइत','सोम','मंगल','बुध','बिही','शुक्र','शनि']
		return days[day-1]

	@staticmethod
	def month(month):
		months = ['बैशाख','जेठ','असार','साउन','भदौ','असोज','कात्तिक','मंसिर','पुस','माघ','फागुन','चैत']
		return months[month-1]

class EnglishChar:
	months = ['Baishakh','Jestha','Ashad','Sharwan','Bhadra','Ashwin','Kartik','Mangsir','Poush','Magh','Falgun','Chaitra']
	days = ['Sunday','Monday','Tuesday','Wednesday','Thursday','Friday','Saturday']
	days_half = ['Sun','Mon','Tue','Wed','Thu','Fri','Sat']
	
	@staticmethod
	def day(day):
		return EnglishChar.days[day-1]

	@staticmethod
	def half_day(day):
		return EnglishChar.days_half[day-1]

	@staticmethod
	def month(month):
		return EnglishChar.months[month-1]


def nepali_to_english_text(text):
	# TODO: optimization

	from .number import NepaliNumber

	# replacing months
	for k,v in MONTHS_MAP.items():
		text = text.replace(k, v)
	
	# replacing days
	for k,v in DAYS_MAP.items():
		text = text.replace(k, v)
	
	# replacing half days
	for k,v in HALF_DAYS_MAP.items():
		text = text.replace(k, v)

	for k,v in AM_PM_MAP.items():
		text = text.replace(k, v)

	return NepaliNumber.revert(text)


#
#
#	Author : Ajesh Sen Thapa
#	Website: www.ajesh.com.np
#
#

import time
import datetime as pythonDateTime
import warnings

from nepali.timezone import now, utc_now
from nepali.utils import to_nepali_timezone

from ._converter import NepaliDateConverter


class formater_class_mixin:
	def get_formater_class(self):
		return self.__class__.get_formater_class()

	@classmethod
	def get_formater_class(cls):
		if not hasattr(cls, '__formater_class__Cache'):
			from ._formarter import NepaliDateTimeFormater
			cls.__formater_class__Cache = NepaliDateTimeFormater
		return cls.__formater_class__Cache

	@classmethod
	def get_strptime_method(cls):
		if not hasattr(cls, '_strptime_method_CACHE'):
			from .parser import strptime
			cls._strptime_method_CACHE = strptime
		return cls._strptime_method_CACHE

class nepalidate(formater_class_mixin):
	def __init__(self, year, month, day) -> None:
		self.__year = year
		self.__month = month
		self.__day = day

		# converting to english date
		year_en, month_en, day_en = NepaliDateConverter.nepali_to_english(year, month, day)
		self.__python_date = pythonDateTime.date(year_en, month_en, day_en)

	def __str__(self):
		return self.strftime_en('%Y-%m-%d')

	def __repr__(self):
		return "<nepalidate> "+str(self)
	
	def to_datetime(self):
		return to_nepali_timezone(pythonDateTime.datetime.combine(self.__python_date, pythonDateTime.time()))

	def to_date(self):
		return self.__python_date

	def to_nepalidatetime(self):
		return nepalidatetime.from_nepali_date(self)

	def to_nepali_datetime(self):
		warnings.warn(
			message="nepalidate.to_nepali_datetime is depreciated and no longer be available in version >= 1.0.0, use nepalidate.to_nepalidatetime instead.", 
			category=DeprecationWarning
		)
		return self.to_nepalidatetime()

	def strftime(self, format):
		NepaliDateTimeFormater = self.get_formater_class()
		formater = NepaliDateTimeFormater(self)
		return formater.get_str(format)

	def strftime_en(self, format):
		NepaliDateTimeFormater = self.get_formater_class()
		formater = NepaliDateTimeFormater(self, english=True)
		return formater.get_str(format)

	
	# operators overloading
	# TODO: add more operators overloading
	def __eq__(self, other):
		""" equal """
		
		if type(other) == self.__class__:
			"""
			NepaliDate object
			"""
			return self.to_date() == other.to_date()

		elif type(other) == pythonDateTime.date:
			"""
			pythonDate object
			"""
			return self.to_date() == to_nepali_timezone(other)
			
		return False

	# static methods
	@classmethod
	def strptime(cls, datetime_str, format):
		nepalidatetime_strptime = cls.get_strptime_method()
		return nepalidatetime_strptime(datetime_str, format=format).date()

	@staticmethod
	def now(*args, **kwargs):
		return nepalidate.today()

	@staticmethod
	def today():
		year, month, day = NepaliDateConverter.current_nepali_date()
		return nepalidate(year, month, day)

	@staticmethod
	def from_datetime(datetime_object):
		return nepalidate.from_date(datetime_object.date())

	@staticmethod
	def from_date(date_object):
		npDate = nepalidate(*NepaliDateConverter.english_to_nepali(date_object.year, date_object.month, date_object.day))
		return npDate

	@staticmethod
	def from_nepalidatetime(datetime_object):
		return datetime_object.date()

	# property

	def weekday(self):
		return self.__python_date.weekday()

	# nepali date properties
	@property
	def year(self):
		return self.__year

	@property
	def month(self):
		return self.__month

	@property
	def day(self):
		return self.__day

	@property
	def week_day(self):
		warnings.warn(
			message="nepalidate.week_day field is depreciated and no longer be available in version >= 1.0.0, use nepalidate.weekday method instead.", 
			category=DeprecationWarning
		)
		return self.weekday()

class nepalitime(pythonDateTime.time):

	def __repr__(self):
		return "<nepalitime> "+str(self)

	# static methods
	
	@staticmethod
	def now(*args, **kwargs):
		dt_now = pythonDateTime.datetime.now()
		return nepalitime(dt_now.hour, dt_now.minute, dt_now.second, dt_now.microsecond)


class nepalidatetime(formater_class_mixin):
	"""
	nepalidatetime
	"""

	def __init__(self, year, month, day, hour=0, minute=0, second=0, microsecond=0):
		self.__npDate = nepalidate(year, month, day)
		self.__npTime = nepalitime(hour, minute, second, microsecond) 

	def __str__(self):
		return str(self.__npDate)+' '+str(self.__npTime)

	def __repr__(self):
		return "<nepalidatetime> "+str(self)


	# operator overloadings

	def __add__(self, other):
		""" addition """

		if type(other) == pythonDateTime.timedelta:
			"""
			timedelta object
			"""
			return nepalidatetime.from_datetime(self.to_datetime() + other)


		return None

	def __sub__(self, other):
		""" substraction """

		if type(other) == self.__class__:
			"""
			nepalidatetime object
			"""
			return self.to_datetime() - other.to_datetime()

		elif type(other) == pythonDateTime.datetime:
			return self.to_datetime() - to_nepali_timezone(other)

		elif type(other) == pythonDateTime.timedelta:
			"""
			timedelta object
			"""
			return nepalidatetime.from_datetime(self.to_datetime() - other)


		return None

	def __lt__(self, other):
		""" less than """

		if type(other) == self.__class__:
			"""
			nepalidatetime object
			"""
			return self.to_datetime() < other.to_datetime()

		elif type(other) == pythonDateTime.datetime:
			"""
			pythonDateTime object
			"""
			return self.to_datetime() < to_nepali_timezone(other)


		return None

	def __le__(self, other):
		""" less than equal """

		if type(other) == self.__class__:
			"""
			nepalidatetime object
			"""
			return self.to_datetime() <= other.to_datetime()

		elif type(other) == pythonDateTime.datetime:
			"""
			pythonDateTime object
			"""
			return self.to_datetime() <= to_nepali_timezone(other)


		return None

	def __eq__(self, other):
		""" equal """
		
		if type(other) == self.__class__:
			"""
			nepalidatetime object
			"""
			return self.to_datetime() == other.to_datetime()

		elif type(other) == pythonDateTime.datetime:
			"""
			pythonDateTime object
			"""
			return self.to_datetime() == to_nepali_timezone(other)

		return False


	def __ne__(self, other):
		""" not equal """

		if type(other) == self.__class__:
			"""
			nepalidatetime object
			"""
			return self.to_datetime() != other.to_datetime()

		elif type(other) == pythonDateTime.datetime:
			"""
			pythonDateTime object
			"""
			return self.to_datetime() != to_nepali_timezone(other)
		
		return True
	
	def __gt__(self, other):
		""" greater than """

		if type(other) == self.__class__:
			"""
			nepalidatetime object
			"""
			return self.to_datetime() > other.to_datetime()

		elif type(other) == pythonDateTime.datetime:
			"""
			pythonDateTime object
			"""
			return self.to_datetime() > to_nepali_timezone(other)

			
		return None
	
	def __ge__(self, other):
		""" greater than equal """

		if type(other) == self.__class__:
			"""
			nepalidatetime object
			"""
			return self.to_datetime() >= other.to_datetime()

		elif type(other) == pythonDateTime.datetime:
			"""
			pythonDateTime object
			"""
			return self.to_datetime() >= to_nepali_timezone(other)

			
		return None 

	# object transformation
	def to_datetime(self):
		return to_nepali_timezone(pythonDateTime.datetime.combine(self.__npDate.to_date(), self.__npTime))

	def to_date(self):
		return self.to_datetime().date()

	def date(self):
		return self.__npDate

	def time(self):
		return self.__npTime

	# string format
	def strftime(self, format):
		NepaliDateTimeFormater = self.get_formater_class()
		formater = NepaliDateTimeFormater(self)
		return formater.get_str(format)

	def strftime_en(self, format):
		NepaliDateTimeFormater = self.get_formater_class()
		formater = NepaliDateTimeFormater(self, english=True)
		return formater.get_str(format)

	# static methods

	@classmethod
	def strptime(cls, datetime_str, format):
		nepalidatetime_strptime = cls.get_strptime_method()
		return nepalidatetime_strptime(datetime_str, format=format)

	@staticmethod
	def from_datetime(dt):
		dt = to_nepali_timezone(dt)
		nd = nepalidate.from_date(dt.date())
		return nepalidatetime(nd.year, nd.month, nd.day, dt.hour, dt.minute, dt.second, dt.microsecond)

	@staticmethod
	def from_date(date_object):
		nepali_date_object = nepalidate.from_date(date_object)
		return nepalidatetime.from_nepali_date(nepali_date_object)

	@staticmethod
	def from_nepali_date(nepali_date_object):
		return nepalidatetime(nepali_date_object.year, nepali_date_object.month, nepali_date_object.day)

	@staticmethod
	def now():
		return nepalidatetime.from_datetime(utc_now())

	# property

	@property
	def year(self):
		return self.__npDate.year

	@property
	def month(self):
		return self.__npDate.month

	@property
	def day(self):
		return self.__npDate.day

	@property
	def week_day(self):
		return self.__npDate.weekday()

	@property
	def hour(self):
		return self.__npTime.hour

	@property
	def minute(self):
		return self.__npTime.minute

	@property
	def second(self):
		return self.__npTime.second


class NepaliDate(nepalidate):
	def __init__(self, *args, **kwargs):
		warnings.warn(
			message="NepaliDate is depreciated and no longer be available in version >= 1.0.0, use nepalidate instead.", 
			category=DeprecationWarning
		)
		super().__init__(*args, **kwargs)

class NepaliTime(nepalidate):
	def __init__(self, *args, **kwargs):
		warnings.warn(
			message="NepaliTime is depreciated and no longer be available in version >= 1.0.0, use nepalitime instead.", 
			category=DeprecationWarning
		)
		super().__init__(*args, **kwargs)

class NepaliDateTime(nepalidate):
	def __init__(self, *args, **kwargs):
		warnings.warn(
			message="NepaliDateTime is depreciated and no longer be available in version >= 1.0.0, use nepalidatetime instead.", 
			category=DeprecationWarning
		)
		super().__init__(*args, **kwargs)

